//
// GDI functions used for printing
//
#include "ratsoft.ch"

#include "Print.ch"
#include "WinApi.ch"
#include "Colors.ch"

// Print info return array
#define DEV_NAME    1
#define DEV_DRIVER  2
#define DEV_PORT    3
#define DEV_TOTAL   3


#define TRANSPARENT 1
#define OPAQUE      2

#define DC_FIELDS           1
#define DC_PAPERS           2
#define DC_PAPERSIZE        3
#define DC_MINEXTENT        4
#define DC_MAXEXTENT        5
#define DC_BINS             6
#define DC_DUPLEX           7
#define DC_SIZE             8
#define DC_EXTRA            9
#define DC_VERSION          10
#define DC_DRIVER           11
#define DC_BINNAMES         12
#define DC_ENUMRESOLUTIONS  13
#define DC_FILEDEPENDENCIES 14
#define DC_TRUETYPE         15
#define DC_PAPERNAMES       16
#define DC_ORIENTATION      17
#define DC_COPIES           18




// Hatches
#define HS_HORIZONTAL       0
#define HS_VERTICAL         1
#define HS_FDIAGONAL        2
#define HS_BDIAGONAL        3
#define HS_CROSS            4
#define HS_DIAGCROSS        5

// Text alignment
#define TA_LEFT                0
#define TA_RIGHT               2
#define TA_CENTER              6


//
static osPrn      ,;   // Current printer object
       osFont     ,;   // Current font
       asMemoLine ,;   // Parsed memo array
       asDevice   ,;
       nsPoint

//
// Configure printing system
//
function pr_Config()

   local aPoint := NIL

   // Set up current device
   pr_GetDevice()

   // Create priter object
   osPrn := TbPrinter():New( "TheBook" )

   // Default background
   SetBkMode( osPrn:hDC, TRANSPARENT )
   //SetBkMode( osPrn:hDC, OPAQUE)

   // Set our mapping mode
   osPrn:SetPixelMode()

   // Create a default font
   pr_CreateFont()

   // Start a page
   osPrn:StartPage()


return( osPrn )

//
// Get device info from win.ini
//
function pr_GetDevice( )

   local lRetval  := TRUE ,;
         cStr     := NIL


   if ( asDevice == NIL )
      asDevice := array( DEV_TOTAL )
   endif

   cStr := GetProfStr( "windows", "device", "" )
   if empty( cStr )
      lRetval := FALSE
   else
      // Parse out the three parts
      asDevice[DEV_NAME  ] := token( cStr, ",", 1 )
      asDevice[DEV_DRIVER] := token( cStr, ",", 2 )
      asDevice[DEV_PORT  ] := token( cStr, ",", 3 )
   endif

return( lRetval )

//
// Return device name
//
function pr_Device()

   local cRetval := asDevice[DEV_NAME  ]

return( cRetval )

//
// Send reset to printer and kill printer object
//
function pr_Reset()

   if !( osPrn == NIL )
      osPrn:End()
   endif

   if !( osFont == NIL )
      osFont:End()
   endif

return( NIL )

//
// Set printer copies
//
function pr_Copies( nCopies )
   osPrn:SetCopies( nCopies )
return( NIL )


//
// Mapping mode to device coordinate functions, including offset correction
//
function row2dc( nNum )
   local nRetval := NIL
   nRetval := ( nNum * osPrn:nVertRes() / osPrn:nVertSize() ) - osPrn:nYoffset
return( max( 0, nRetval ) )

function col2dc( nNum )
  local nRetval := NIL
  nRetval := ( nNum * osPrn:nHorzRes()/osPrn:nHorzSize() ) - osPrn:nXoffset
return( max( 0, nRetval ) )

//
// Misc conversion
//
function pix2mm( nPx )
return( nPx / osPrn:nLogPixelX * 25.4 )

function pix2in( nPx )
return( nPx / osPrn:nLogPixelX )

function pt2mm( nPt )
return( nPt / 72 * 25.4 )

//
// Get paper size
//
function pr_PageWidth()
return( osPrn:nHorzSize + ( 2 * pix2mm( osPrn:nXoffset ) ) )

function pr_PageHeight()
return( osPrn:nVertSize + ( 2 * pix2mm( osPrn:nYoffset ) ) )


//
// Return length of string in mm
//
function pr_Len( cTxt )

   local nPixels := GetTextWidth( osPrn:hDC, cTxt, osFont:hFont ) ,;
         nRetval := NIL

   nRetval := pix2mm( nPixels )

return( nRetval )

//
// Create a font for use by pr_Say()
//
function pr_CreateFont( cFont, nPoint, lBold, lItalic )

   local nWidth  := NIL ,;
         nHeight := NIL

   DEFAULT( cFont   , "C", "Times New Roman")
   DEFAULT( nPoint  , "N", 12 )
   DEFAULT( lBold   , "L", FALSE )
   DEFAULT( lItalic , "L", FALSE )

   // Release any old fonts
   if !( osFont == NIL )
      osFont:End()
   endif

   // Map font point to pixels
   nsPoint := nPoint

   // This requirement was removed in FW 1.9.1
   //nHeight := nPoint * osPrn:nLogPixelY() / 72
   // Simply pass the point height
   nHeight := nPoint

   osFont := TFont():New( cFont, nWidth, nHeight,;
                          FALSE   ,; // <.from.>,;
                          lBold   ,; // [<.bold.>],
                          NIL     ,; // <nEscapement>,
                          NIL     ,;
                          NIL     ,; // <nWeight>       ,
                          lItalic ,; // [<.italic.>]    ,;
                          FALSE   ,; // [<.underline.>] ,
                          FALSE   ,; // Strikeout
                          NIL     ,; // nCharset
                          NIL     ,; // nOutPrecision
                          NIL     ,; // nClipPrecision
                          NIL     ,; // nQuality,
                          osPrn )    // [<oDevice>] )

return( NIL )

//
// Set font
//
function pr_SetFont( cFont, nPoint, lBold, lItalic )

   local nHeight := NIL


    DEFAULT( cFont   , "C", "Times New Roman")
    DEFAULT( nPoint  , "N", 12 )
    DEFAULT( lBold   , "L", FALSE )
    DEFAULT( lItalic , "L", FALSE )

    if !( nsPoint == nPoint           ) .or. ;
       !( cFont   == osFont:cFaceName ) .or. ;
       !( lBold   == osFont:lBold     ) .or. ;
       !( lItalic == osFont:lItalic   )

       pr_CreateFont( cFont, nPoint, lBold, lItalic )

    endif

return( NIL )


//
// Say text string to printer
//
function pr_Say( nRow, nCol, cTxt, cJust )

   local nLen        := NIL

   DEFAULT( cJust, "C", PR_JUST_LEFT )

   // Correct coords for Mapping Mode
   nRow := row2dc( nRow )

   // Adjust for justification
   do case
      case ( cJust == PR_JUST_LEFT   )
         SetTextAlign( osPrn:hDC, TA_LEFT )

      case ( cJust == PR_JUST_RIGHT  )
         SetTextAlign( osPrn:hDC, TA_RIGHT )

      case ( cJust == PR_JUST_CENTRE )
         SetTextAlign( osPrn:hDC, TA_CENTER )

   endcase
   nCol     := col2dc( nCol )

   // And say it
   osPrn:Say( nRow, nCol, cTxt, osFont )

return( NIL )



//
// Draw a line
//
function pr_Line( nRow, nCol, nLen, oPen, nDirection )

   local nBottom := NIL ,;
         nRight  := NIL

   DEFAULT( oPen        , "O", TPen():New() )
   DEFAULT( nDirection  , "N", LINE_HORIZONTAL )


   // Check direction
   do case
      case ( nDirection == LINE_HORIZONTAL )
         nBottom := nRow
         nRight  := ( nCol + nLen )

      case ( nDirection == LINE_VERTICAL )
         nBottom := ( nRow + nLen )
         nRight  := nCol

      case ( nDirection == LINE_DIAGONAL )
         nBottom := ( nRow + nLen )
         nRight  := ( nCol + nLen )

   endcase

   // Correct coords for Mapping Mode
   nRow     := row2dc( nRow    )
   nCol     := col2dc( nCol    )
   nBottom  := row2dc( nBottom )
   nRight   := col2dc( nRight  )


   // Draw it
   osPrn:Line( nRow, nCol, nBottom, nRight, oPen )

return( NIL )

//
// Draw a box
//
function pr_Box( nTop, nLeft, nBottom, nRight, nLType ) // , nGrey )

   local lRect    := FALSE ,;
         nStyle   := NIL ,;
         nWidth   := NIL ,;
         nColor   := NIL ,;
         oPen     := NIL ,;
         hBMP     := NIL ,;
         oBrush   := NIL ,;
         hBrush   := NIL

   DEFAULT( nLType, "N", LINE_SINGLE   )


   // Correct for mapping mode
   nTop     := row2dc( nTop    )
   nLeft    := col2dc( nLeft   )
   nBottom  := row2dc( nBottom )
   nRight   := col2dc( nRight  )

   // Could be a little more imaginative here..
   nWidth := 1

   // And here
   nColor := CLR_BLACK

   //
   do case
      case ( nLType == LINE_NONE   )
         nStyle := PS_NULL

      case ( nLType == LINE_SINGLE )
         nStyle := PS_SOLID

      case ( nLType == LINE_DOTTED )
         nStyle := PS_DOT

      case ( nLType == LINE_HAIR   )
         nStyle := PS_SOLID

      case ( nLType == LINE_DASHED )
         nStyle := PS_DASH

      case ( nLType == LINE_THIN   )
         nStyle := PS_SOLID

      case ( nLType == LINE_DOUBLE )
         nStyle := PS_SOLID

   endcase
   oPen  := TPen():New( nStyle, nWidth, nColor, osPrn )

   // Now draw the box
   osPrn:Line( nTop, nLeft , nTop, nRight , oPen )
   osPrn:Line( nTop, nRight, nBottom, nRight, oPen )

   osPrn:Line( nTop, nLeft, nBottom, nLeft, oPen )
   osPrn:Line( nBottom, nLeft, nBottom, nRight, oPen )

   // Clean up
   oPen:End()

return( NIL )

//
// Shaded rectangle
//
function pr_Rectangle( nTop, nLeft, nBottom, nRight, nGrey )

   local nColor := CLR_GRAY ,;
         hBrush := NIL      ,;
         oBrush := NIL

   DEFAULT( nGrey, "N", SHADE_LIGHT )

   // Correct for mapping mode
   nTop     := row2dc( nTop    )
   nLeft    := col2dc( nLeft   )
   nBottom  := row2dc( nBottom )
   nRight   := col2dc( nRight  )

   // Determine shade requested
   do case
      case ( nGrey == SHADE_LIGHT  )
         oBrush := TBrush():New( , RGB( 245, 245, 245 ) )
         //hBrush   := CreateSolidBrush( RGB( 245, 245, 245 ) )

      case ( nGrey == SHADE_MEDIUM )
         oBrush := TBrush():New( , RGB( 200, 200, 200 ) )
         //hBrush   := CreateSolidBrush( RGB( 200, 200, 200 ) )

      case ( nGrey == SHADE_DARK   )
         oBrush := TBrush():New( , RGB( 125, 125, 125 ) )
         //hBrush   := CreateSolidBrush( RGB( 125, 125, 125 ) )

   endcase


   //osPrn:FillRectangle( { nTop, nLeft, nBottom, nRight }, hBrush )
   osPrn:FillRect( { nTop, nLeft, nBottom, nRight }, oBrush )

   DeleteObject( hBrush )

return( NIL )


//
// Draw a bitmap to fit nWidth, nHeight
//
function pr_BitMap( nRow, nCol, xBitmap, nWidth, nHeight, nRc, lHCtr, lVCtr )

   DEFAULT( nRc, "N", 0 )

   // Correct coords for Mapping Mode
   nRow     := row2dc( nRow    )
   nCol     := col2dc( nCol   )

   osPrn:DrawBitMap( nRow, nCol, xBitmap, nWidth, nHeight, nRc, lHCtr, lVCtr )

return( NIL )

//
// Spit out page
//
function pr_Eject( lFinal )

   DEFAULT( lFinal, "L", FALSE )

   osPrn:EndPage()
   if !lFinal
      osPrn:StartPage()
   endif

return( NIL )

//
// Start and End a page
//
function pr_StartPage()
   osPrn:StartPage()
return( NIL )

function pr_EndPage()
   osPrn:EndPage()
return( NIL )

//
// Set/reset paper orientation. Default is portrait
//
function pr_Landscape( lParm )

   static lsLand := FALSE
   local lRetval := lsLand

   if ( lParm == NIL )
      lRetval := lsLand
   else
      if lParm
         osPrn:SetLandscape()
      else
         osPrn:SetPortrait()
      endif
      lsLand := lParm
   endif

return( lRetval )

//
//  Count number of lines in a string based on width
//  And at the same time accumulate lines in master array asMemoLine
//
function pr_JustCount( cTxt, nWidth, cDelim )

   local nLines   := 0     ,;
         cLine    := NIL   ,;
         cTmp     := NIL   ,;
         cTok     := NIL   ,;
         aPara    := {}    ,;
         cPara    := ""    ,;
         i        := NIL   ,;
         j        := NIL   ,;
         nPara    := NIL   ,;
         nSpace   := NIL   ,;
         nTokens  := NIL   ,;
         nLen     := NIL   ,;
         nPos     := NIL

   DEFAULT( cDelim, "C", " " )

   // Reset memoline array
   asMemoLine := {}

   // Replace all CRLF with one char
   cTxt := strtrani( cTxt, CRLF, "|" )

   // Process each paragraph
   nPara := numtoken( cTxt, "|" )
   for i := 1 to nPara

      cPara := token( cTxt, "|", i )

      // And wrap tokens (words) within the para
      cLine    := ""
      nTokens  := numtoken( cPara, " " )
      for j := 1 to nTokens

         cTok  := token( cPara, " ", j )

         // See if the current line would fit
         if ( pr_Len( cLine +" "+ cTok ) < nWidth )
            // Yes, it fits so keep going
            cLine += ( iif( empty(cLine), "", " ") + cTok )
         else
            // No, cLine is maximum
            aadd( asMemoLine, cLine )
            cLine := cTok
         endif

      next // j word

      // Any remainder from the paragraph as new line
      if !empty( cLine )
         aadd( asMemoLine, cLine )
         cLine := ""
      endif

   next // i paragraph

   // Retun value
   nLines := len( asMemoLine )

return( nLines )

//
// Return a specific line from the memo array
//
function pr_JustLine( nLine )

return( asMemoLine[nLine] )


//
// Shell function for pr_JustCount() which returns the actual lines
// and not just the line count
//
function pr_JustACount( cTxt, nWidth )

   pr_JustCount( cTxt, nWidth )

return( aclone( asMemoLine ) )


//
// Printer DUPLEX capabilities
//
function pr_IsDuplex()

   local nDuplex := NIL ,;
         lRetval := NIL

   nDuplex := DeviceCapa( asDevice[DEV_NAME], DC_DUPLEX )

   if ( nDuplex == 1 )
      lRetval := TRUE
   else
      lRetval := FALSE
   endif

return( lRetval )

function pr_Duplex( nMode )
   if pr_Landscape()
   else
   endif
return( NIL )

// function pr_Test()
//
//    local cTmp := space(50)
//
//    DeviceCapa( asDevice[DEV_NAME], DC_PAPERNAMES, @cTmp )
//
//    do case
//       case ( valtype(cTmp) == "C" )
//          MsgInfo( "Papernames are "+cTmp )
//       case ( valtype(cTmp) == "N" )
//          MsgInfo( "Papernames are "+ltrim(str(cTmp)) )
//       otherwise
//          MsgInfo( "Incorrect return value from DeviceCapa" )
//    endcase
//
// return( NIL )


//
// Application compatibility stub functions
//
function pr_ChangeFont( cFont, nPoint )
return( TRUE )

function pr_Initialise( cName )
return( NIL )

function pr_ManualFeed()
return( NIL )

function pr_PJLMsg()
return( NIL )

function pr_MacroCreate()
return( NIL )

function pr_FntInit()
return( NIL )

function pr_MacroUse()
return( NIL )

function pr_Set()
return( NIL )


